#define _LARGEFILE64_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <time.h>
#include <errno.h>
#include <sys/select.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#include <ctype.h>
#include <pwd.h>

extern "C" {
#include "error.h"
#include "apache_log.h"
}
#include "pl.h"
#include "utils.h"
#include "gen.h"
#include "rss.h"

char *state_str[4] = { "OK", "WARNING", "CRITICAL", "UNKNOWN" };
char *state_type_str[2] = { "SOFT", "HARD" };

int setup_listen_socket(int port)
{
	int     reuse_addr = 1;
        struct sockaddr sock_name;
        struct sockaddr_in *sock_name_in = (sockaddr_in *)&sock_name;
        socklen_t sock_name_len = sizeof(sock_name);

        /* create socket */
        int handle = socket(AF_INET, SOCK_STREAM, 0);
        if (handle == -1)
                error_exit("failed to create socket\n");

	if (setsockopt(handle, SOL_SOCKET, SO_REUSEADDR, (char *)&reuse_addr, sizeof(reuse_addr)) == -1)
		error_exit("setsockopt(SO_REUSEADDR)");

        /* set fd to non-blocking */
        if (fcntl(handle, F_SETFL, O_NONBLOCK) == -1)
                error_exit("failed to set socket to nonblocking\n");

        /* bind socket to port 33333 */
	memset(&sock_name, 0x00, sizeof(sock_name));
        sock_name_in -> sin_port = htons(port);
        if (bind(handle, &sock_name, sock_name_len))
                error_exit("cannot setup socket (bind)");

        /* setup listen queue */
        if (listen(handle, 5) == -1)
                error_exit("cannot setup socket (listen)");

	return handle;
}

char * escape_http(char *in)
{
	int len = strlen(in);
	int out_offset = 0;
	int loop;
	char *out = (char *)mymalloc(len * 3 + 1, "escape_http");

	for(loop=0; loop<len; loop++)
	{
		if (isalnum(in[loop]) && in[loop] != '-')
		{
			out[out_offset++] = in[loop];
		}
		else if (in[loop] == ' ')
			out[out_offset++] = '+';
		else
		{
			out_offset += sprintf(&out[out_offset], "%%%02x", in[loop]);
		}
	}
	out[out_offset] = 0x00;

	return out;
}

char * gen_rss_string(struct stats *pstats, int n_stats, int all, int list_all_problems, int always_notify, int also_acknowledged, int hide_ok, char *nagios_cgibin)
{
	int loop;
	char *rss_str = NULL;
	int n_added = 0;

	/* RSS header */
	rss_str = rss_header("Nagios status RSS feed",			/* title */
			     "http://www.vanheusden.com/nagserv/",	/* link */
			     "Nagios status RSS feed",			/* description */
			     "folkert@vanheusden.com",			/* managing editor */
			     "folkert@vanheusden.com",			/* webmaster */
			     "nagserv v" VERSION ", (C) 2005-2006 by folkert@vanheusden.com");	/* generator */

	/* RSS item for each problem */
	for(loop=0; loop<n_stats; loop++)
	{
		if (should_i_show_entry(pstats, n_stats, loop, list_all_problems, always_notify, also_acknowledged, hide_ok) || all == 1)
		{
			char title[4096];
			char link [4096];

			/* generate title */
			if (pstats[loop].type == TYPE_SERVICE)
			{
				snprintf(title, sizeof(title), "SERVICE: %s @ %s = %s", pstats[loop].service_description,
											pstats[loop].host_name,
											state_str[pstats[loop].current_state]);
			}
			else
			{
				snprintf(title, sizeof(title), "HOST: %s = %s", pstats[loop].host_name,
										state_str[pstats[loop].current_state]);
			}

			/* generate link */
			if (!nagios_cgibin)
				strncpy(link, "http://localhost/nagios/", sizeof(link));
			else
			{
				if (pstats[loop].type == TYPE_HOST)
				{
					snprintf(link, sizeof(link), "%s/extinfo.cgi?type=1&host=%s", nagios_cgibin, pstats[loop].host_name);
				}
				else if (pstats[loop].type == TYPE_SERVICE)
				{
					snprintf(link, sizeof(link), "%s/extinfo.cgi?type=2&host=%s&service=%s", nagios_cgibin, pstats[loop].host_name, pstats[loop].service_description);
				}
			}

			/* add item */
			rss_item(&rss_str, title, link, pstats[loop].plugin_output, pstats[loop].last_state_change);

			n_added++;
		}
	}

	if (n_added == 0)
	{
		char *urls[] = { "http://www.vanheusden.com/multitail/", "http://www.vanheusden.com/pgp.php",
			"http://www.vanheusden.com/httping/",   "http://www.unixexpert.nl/"
		};

		rss_item(&rss_str, "No events - all services are fine!", /* title */
				   urls[rand() % 4],			 /* link */
				   "No systems are reporting problems.", /* description */
				   time(NULL));				 /* timestamp */
	}

	rss_tail(&rss_str);

	return rss_str;
}

char * gen_nagios_1_0_string(struct stats *pstats, int n_stats, int suggested_alloc_size, int all, int list_all_problems, int always_notify, int also_acknowledged, int hide_ok)
{
	char *status_str = (char *)mymalloc(suggested_alloc_size, "status_str");
	int status_str_len = 0;

	for(int loop=0; loop<n_stats; loop++)
	{
		if (should_i_show_entry(pstats, n_stats, loop, list_all_problems, always_notify, also_acknowledged, hide_ok) || all == 1)
		{
			/*
			 * All Host Lines:
			 * [Time of last update] HOST;
			 * Host Name (string);
			 * Status (OK/DOWN/UNREACHABLE);
			 * Last Check Time (long time);
			 * Last State Change (long time);
			 * Acknowledged (0/1);
			 * Time Up (long time);
			 * Time Down (long time);
			 * Time Unreachable (long time);
			 * Last Notification Time (long time);
			 * Current Notification Number (#);
			 * Notifications Enabled (0/1);
			 * Event Handlers Enabled (0/1);
			 * Checks Enabled (0/1);
			 * Flap Detection Enabled (0/1);
			 * Host is Flapping (0/1);
			 * Percent State Change (###.##);
			 * Scheduled downtime depth (#);
			 * Failure Prediction Enabled (0/1);
			 * Process Performance Data(0/1);
			 * Plugin Output (string)
			 *
			 * Service Lines:
			 * [Time of last update] SERVICE;
			 * Host Name (string);
			 * Service Description (string);
			 * Status (OK/WARNING/CRITICAL/UNKNOWN);
			 * Retry number (#/#);
			 * State Type (SOFT/HARD);
			 * Check type (ACTIVE/PASSIVE);
			 * Last check time (long time);
			 * Next check time (long time);
			 * Checks enabled (0/1);
			 * Accept Passive Checks (0/1);
			 * Event Handlers Enabled (0/1);
			 * Last state change (long time);
			 * Problem acknowledged (0/1);
			 * Last Hard State (OK/WARNING/CRITICAL/UNKNOWN);
			 * Time OK (long time);
			 * Time Unknown (long time);
			 * Time Warning (long time);
			 * Time Critical (long time);
			 * Last Notification Time (long time);
			 * Current Notification Number (#);
			 * Notifications Enabled (-1/1);
			 * Latency (#);
			 * Execution Time (#);
			 * Flap Detection Enabled (0/1);
			 * Service is Flapping (0/1);
			 * Percent State Change (###.##);
			 * Scheduled Downtime Depth (#);
			 * Failure Prediction Enabled (0/1);
			 * Process Performance Date (0/1);
			 * Obsess Over Service (0/1);
			 * Plugin Output (string)
			 */
			if (pstats[loop].type == TYPE_HOST)
				status_str_len += sprintf(&status_str[status_str_len],
						"[%d] HOST;%s;%s;%d;%d;%d;%d;%d;%d;%d;%f;%d;%d;%d;%d;%d;%f;%f;%d;%d;%s;\n",
						//  1       2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 18 19 20 21
						(int)pstats[loop].last_update,		// 1
						pstats[loop].host_name,			// 2
						state_str[pstats[loop].current_state],	// 3
						(int)pstats[loop].last_check,		// 4
						(int)pstats[loop].last_state_change,		// 5
						pstats[loop].problem_has_been_acknowledged, // 6
						pstats[loop].last_time_up,		// 7
						pstats[loop].last_time_down,		// 8
						pstats[loop].last_time_unreachable,	// 9
						pstats[loop].last_notification,		// 10
						pstats[loop].current_notification_number, // 11
						pstats[loop].notifications_enabled,	// 12
						pstats[loop].event_handler_enabled,	// 13
						pstats[loop].active_checks_enabled,	// 14
						pstats[loop].flap_detection_enabled,	// 15
						pstats[loop].is_flapping,		// 16
						pstats[loop].percent_state_change,	// 17
						pstats[loop].scheduled_downtime_depth,	// 18
						pstats[loop].failure_prediction_enabled,// 19
						pstats[loop].process_performance_data,	// 20
						pstats[loop].plugin_output		// 21
							);
			else if (pstats[loop].type == TYPE_SERVICE)
				status_str_len += sprintf(&status_str[status_str_len], 
						"[%d] SERVICE;%s;%s;%s;1/1;%s;%d;%d;%s;%d;%d;%d;%d;%d;%s;%d;%d;%d;%d;%d;%f;%d;%f;%f;%d;%d;%f;%f;%d;%d;%d;%s\n",
						//  1           2 3  4      5  6  7  8  9  10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31
						(int)pstats[loop].last_update,		// 1
						pstats[loop].host_name,			// 2
						pstats[loop].service_description,	// 3
						state_str[pstats[loop].current_state],	// 4
						state_type_str[pstats[loop].state_type],// 5
						(int)pstats[loop].last_check,		// 6
						(int)pstats[loop].next_check,		// 7
						pstats[loop].check_type?"ACTIVE":"PASSIVE", // 8
						pstats[loop].active_checks_enabled,	// 9
						pstats[loop].passive_checks_enabled,	// 10
						pstats[loop].event_handler_enabled,	// 11
						(int)pstats[loop].last_state_change,		// 12
						pstats[loop].problem_has_been_acknowledged, // 13
						state_str[pstats[loop].last_hard_state], // 14
						pstats[loop].last_time_ok,		// 15
						pstats[loop].last_time_unknown,		// 16
						pstats[loop].last_time_warning,		// 17
						pstats[loop].last_time_critical,	// 18
						pstats[loop].last_notification,		// 19
						pstats[loop].current_notification_number, // 20
						pstats[loop].notifications_enabled,	// 21
						pstats[loop].check_latency,		// 22
						pstats[loop].check_execution_time,	// 23
						pstats[loop].flap_detection_enabled,	// 24
						pstats[loop].is_flapping,		// 25
						pstats[loop].percent_state_change,	// 26
						pstats[loop].scheduled_downtime_depth,	// 27
						pstats[loop].failure_prediction_enabled, // 28
						pstats[loop].process_performance_data,	// 29
						pstats[loop].obsess_over_service,	// 30
						pstats[loop].plugin_output		// 31
							);
			else
				error_exit("internal error: %d", pstats[loop].type);

			if (status_str_len > suggested_alloc_size)
				error_exit("unexpected output string size (%d > %d)", status_str_len, suggested_alloc_size);
		}
	}
	status_str[status_str_len] = 0x00;

	return status_str;
}

int WRITE_str(int fd, char *string)
{
	int len = strlen(string);

	return WRITE(fd, string, len);
}

void do_client(int server_fd, int get_http, char *str, time_t last_changed, char *logfile)
{
	int client_fd;
	struct sockaddr addr;
	socklen_t addrlen = sizeof(addr);

	client_fd = accept(server_fd, &addr, &addrlen);
	if (client_fd == -1)
	{
		syslog(LOG_WARNING, "accept() failed: %m");
	}
	else
	{
		pid_t child_pid = fork();

		if (child_pid == -1)
			error_exit("fork() failed");
		else if (child_pid == 0)
		{
			int body = 1;
			int body_size = strlen(str);

			if (get_http)
			{
				char *dummy;
				struct tm *pgmtime = gmtime(&last_changed);
				char recv_buffer[4096], send_buffer[4096];

				/* receive request headers */
				for(;;)
				{
					if (read(client_fd, recv_buffer, sizeof(recv_buffer)) >= 0)
						break;

					if (errno != EINTR && errno != EAGAIN)
						exit(0);
				}
				if (strncmp(recv_buffer, "HEAD ", 5) == 0)
					body_size = body = 0;

				dummy = strchr(recv_buffer, '\r');
				if (!dummy)
					dummy = strchr(recv_buffer, '\n');
				if (dummy)
				{
					*dummy = 0x00;
					(void)apachelog(logfile, ((sockaddr_in *)&addr) -> sin_addr, time(NULL), 200, (off64_t)body_size, recv_buffer, "" /* referer */, "" /* browser */);
				}

				/* send reply headers */
				WRITE_str(client_fd, "HTTP/1.0 200 OK\r\n");
				/* Last-Modified: Wed, 31 Mar 2004 18:41:41 GMT */
				snprintf(send_buffer, sizeof(send_buffer), "Last-Modified: %s, %d %s %d %02d:%02d:%02d GMT\r\n",
									   wday[pgmtime -> tm_wday],
									   pgmtime -> tm_mday,
									   month[pgmtime -> tm_mon],
									   pgmtime -> tm_year + 1900,
									   pgmtime -> tm_hour,
									   pgmtime -> tm_min,
									   pgmtime -> tm_sec);
				WRITE_str(client_fd, send_buffer);
				WRITE_str(client_fd, "Content-Type: text/xml\r\n");
				WRITE_str(client_fd, "Connection: close\r\n\r\n");
			}

			if (body)
				WRITE(client_fd, str, body_size);

			close(client_fd);

			exit(0);
		}

		close(client_fd);
	}
}

void start_proc(char *do_exec, char *par)
{
	pid_t child = fork();

	if (child == -1)
		error_exit("failed to start process '%s'", do_exec);
	else if (child == 0)
	{
		if (-1 == execlp("/bin/sh", "/bin/sh", "-c", do_exec, par, (void *)NULL)) error_exit("execlp of %s failed\n", do_exec);
	}
}

void load_statuslog(char *statuslog, char is_file, char is_20_format, int prev_n_elements_in_stats_array, struct stats **pstats, int *n_stats)
{
	char reload = 0;

	for(;;)
	{
		int fd;

		/* open file or connection to nagios status socket */
		if (is_file == 1)     /* file */
			fd = open64(statuslog, O_RDONLY);
		else
			fd = connect_to(statuslog);
		if (fd == -1)
		{
			syslog(LOG_INFO, "Failed to open nagios status-log at %s", statuslog);
			printf("Failed to open nagios status-log at %s\n", statuslog);
			sleep(1);
			continue;
		}

		/* file/socket open, load data */
		if (is_20_format)
			parse_2_0_statuslog(fd, pstats, n_stats);
		else
			parse_1_0_statuslog(fd, pstats, n_stats);

		close(fd);

		if (*n_stats == prev_n_elements_in_stats_array)
		{
			break;
		}

		syslog(LOG_INFO, "Number of elements in status-log is different (%d) from previous run (%d), retrying", *n_stats, prev_n_elements_in_stats_array);
		printf("Number of elements in status-log is different (%d) from previous run (%d), retrying\n", *n_stats, prev_n_elements_in_stats_array);
		reload = 1;
		sleep(1);

		/* remember the current number of elements for this load */
		prev_n_elements_in_stats_array = *n_stats;
		/* and free up - it'll be reloaded */
		free_stats_array(*pstats, *n_stats);
	}

	if (reload)
	{
		syslog(LOG_INFO, "Size of status.log is stable (%d elements), continuing", *n_stats);
		printf("Size of status.log is stable (%d elements), continuing\n", *n_stats);
	}
}

void version(void)
{
	fprintf(stderr, "nagserv v" VERSION ", (C) 2005-2006 folkert@vanheusden.com\n");
}

void usage(void)
{
	printf("-f x     path and filename to the nagios status.log (the file\n");
	printf("         selected in nagios.cfg with 'status_file')\n");
	printf("-p x     portnumber to listen on (for nagios 1.x output)\n");
	printf("-r x     portnumber to listen on (for RSS feed)\n");
	printf("-L x     log RSS requests in apache logformat to file x\n");
	printf("-a       do not filter, send all records to the client\n");
	printf("-z x     when the status has changed, execute command 'x'\n");
	printf("-l x     set URL(!) to your nagios cgi-bin directory (for RSS feed)\n");
	printf("-t x     checks to see if nagios is still running. comma-seperated list\n");
	printf("         (without spaces!) with the following elements:\n");
	printf("         max_time_last_host_update, max_time_oldest_host_update,\n");
	printf("         max_time_last_host_check, max_time_oldest_host_check,\n");
	printf("         max_time_last_service_check, max_time_oldest_service_check,\n");
	printf("         max_time_oldest_next_service_check\n");
	printf("-T x     script to use for the -t check\n");
	printf("-F       host:port for retrieving status.log\n");
	printf("-x       status.log is in nagios 1.0 format\n");
	printf("-X       status.log is in nagios 2.0 format\n");
	printf("-i       how often to check if the status.log got changed: use only\n");
	printf("         when you use the '-z' parameter\n");
	printf("-Z       do not fork into the background\n");
	printf("-u x     run as user 'x'\n");
	printf("-P x     write pid to file x\n");
	printf("-V       show version number and exit\n");
	printf("-h       this help\n");
}

int main(int argc, char *argv[])
{
	char *statuslog = "/usr/local/nagios/var/status.log";
	int nag1_port = 0;
	char list_all_problems = 0;
	char always_notify = 0;
	char also_acknowledged = 0;
	char hide_ok = 1;
	int all = 0;
	int do_not_fork = 0;
	int sw;
	int rss_port = 0;
	char *rss_str = NULL;
	char *nag1_str = NULL;
	int rss_fd = -1, nag1_fd = -1;
	char *do_exec = NULL;
	time_t last_file_change = (time_t)0;
	char *nagios_cgibin = NULL;
	int check_interval = -1;
	char *rss_logfile = NULL;
	char is_20_format = 0;
	char is_file = 1;
	char *max_age_last_check_script = NULL;
	int max_time_last_host_update = -1, max_time_oldest_host_update = -1, max_time_last_host_check = -1, max_time_oldest_host_check = -1, max_time_last_service_check = -1, max_time_oldest_service_check = -1, max_time_oldest_next_service_check = -1;
	char first_run = 1;
	int prev_n_elements_in_stats_array = 0;
	char *runas = NULL;
	char *pidfile = NULL;

	srand(time(NULL));

	while ((sw = getopt(argc, argv, "P:ZL:i:f:p:r:z:l:axXF:hVt:T:u:")) != EOF)
	{
		switch(sw)
		{
			case 'P':
				pidfile = optarg;
				break;

			case 'u':
				runas = optarg;
				break;

			case 't':
				{
					char *dummy = optarg;

					max_time_last_host_update = atoi(dummy);
					dummy = strchr(dummy, ',') + 1;

					max_time_oldest_host_update = atoi(dummy);
					dummy = strchr(dummy, ',') + 1;

					max_time_last_host_check = atoi(dummy);
					dummy = strchr(dummy, ',') + 1;

					max_time_oldest_host_check = atoi(dummy);
					dummy = strchr(dummy, ',') + 1;

					max_time_last_service_check = atoi(dummy);
					dummy = strchr(dummy, ',') + 1;

					max_time_oldest_service_check = atoi(dummy);
					dummy = strchr(dummy, ',') + 1;

					max_time_oldest_next_service_check = atoi(dummy);
					break;
				}

			case 'T':
				max_age_last_check_script = mystrdup(optarg);
				break;

			case 'L':
				rss_logfile = mystrdup(optarg);
				break;

			case 'f':
				statuslog = mystrdup(optarg);
				is_file = 1;
				break;

			case 'F':
				statuslog = mystrdup(optarg);
				is_file = 0;
				break;

			case 'x':
				is_20_format = 0;
				break;

			case 'X':
				is_20_format = 1;
				break;

			case 'p':
				nag1_port = atoi(optarg);
				if (nag1_port < 1 || nag1_port > 65535 || nag1_port == rss_port)
					error_exit("invalid portnumber given");
				break;

			case 'r':
				rss_port = atoi(optarg);
				if (rss_port < 1 || rss_port > 65535 || rss_port == nag1_port)
					error_exit("invalid portnumber given");
				break;

			case 'l':
				nagios_cgibin = mystrdup(optarg);
				break;

			case 'a':
				all = 1;
				break;

			case 'z':
				do_exec = mystrdup(optarg);
				break;

			case 'i':
				check_interval = atoi(optarg);
				break;

			case 'Z':
				do_not_fork = 1;
				break;

			case 'V':
				version();
				return 0;

			case 'h':
				usage();
				return 0;

			default:
				usage();
				return 1;
		}
	}

	if (nag1_port == 0 && rss_port == 0 && max_age_last_check_script == NULL)
		error_exit("no service selected: use -p, -r or -t/-T");

	if ((do_exec != NULL || nagios_cgibin != NULL) && rss_port == 0)
		error_exit("-z or -l given while no RSS port was given");

	if (rss_port == 0 && rss_logfile != NULL)
		error_exit("no rss port selected while enabling rss requests logfile");

	signal(SIGCHLD, SIG_IGN);

	if (runas)
	{
		struct passwd *p = getpwnam(runas);
		if (p)
		{
			if (setgid(p -> pw_gid) == -1)
				error_exit("setgid(%d) failed", p -> pw_gid);

			if (setuid(p -> pw_uid) == -1)
				error_exit("setuid(%d) failed", p -> pw_uid);
		}
		else
		{
			error_exit("User '%s' not found.", runas);
		}
	}

	if (!do_not_fork)
	{
		if (daemon(0, 0) == -1)
			error_exit("failed becoming daemon process");
	}

	if (pidfile)
		write_pidfile(pidfile);

	if (rss_port != 0)
	{
		struct sockaddr addr;
		socklen_t addrlen = sizeof(addr);
		rss_fd = setup_listen_socket(rss_port);
		close(accept(rss_fd, &addr, &addrlen));
	}

	if (nag1_port != 0)
	{
		struct sockaddr addr;
		socklen_t addrlen = sizeof(addr);
		nag1_fd = setup_listen_socket(nag1_port);
		close(accept(rss_fd, &addr, &addrlen));
	}

	for(;;)
	{
		fd_set rfds;
		time_t cur_last_file_change = (time_t)0;
		struct timeval tv;

		FD_ZERO(&rfds);
		if (rss_fd != -1)
			FD_SET(rss_fd, &rfds);
		if (nag1_fd != -1)
			FD_SET(nag1_fd, &rfds);

		if (!first_run)
		{
			tv.tv_usec = 0;
			if (check_interval != -1)
				tv.tv_sec = check_interval;
			else
				tv.tv_sec = 2000000;

			if (select(max(rss_fd, nag1_fd) + 1, &rfds, NULL, NULL, &tv) == -1)
			{
				if (errno != EAGAIN && errno != EINTR)
					error_exit("select failed");

				continue;
			}
		}
		else
		{
			first_run = 0;
		}

		if (is_file)
			cur_last_file_change = get_filechanged(statuslog);

		if (cur_last_file_change != last_file_change || is_file == 0)
		{
			struct stats *pstats = NULL;
			int n_stats = 0;
			int file_size = -1;
			char *prss_str = rss_str, *pnag1_str = nag1_str;

			load_statuslog(statuslog, is_file, is_20_format, prev_n_elements_in_stats_array, &pstats, &n_stats);
			prev_n_elements_in_stats_array = n_stats;

			if (is_file)
				file_size = get_filesize(statuslog);

			if (max_age_last_check_script != NULL)
			{
				char *message = NULL;

				if (check_max_age_last_check(pstats, n_stats, max_time_last_host_update, max_time_oldest_host_update, max_time_last_host_check, max_time_oldest_host_check, max_time_last_service_check, max_time_oldest_service_check, max_time_oldest_next_service_check, &message) == -1)
				{
					printf("down: %s\n", message);
					start_proc(max_age_last_check_script, message);
					free(message);
				}
			}

			sort_stats_array(pstats, n_stats);

			if (rss_fd != -1)
				rss_str = gen_rss_string(pstats, n_stats, all, list_all_problems, always_notify, also_acknowledged, hide_ok, nagios_cgibin);
			if (nag1_fd != -1)
				nag1_str = gen_nagios_1_0_string(pstats, n_stats, file_size != -1 ? file_size : n_stats * 1024, all, list_all_problems, always_notify, also_acknowledged, hide_ok);

			if (do_exec)
			{
				if (prss_str == NULL || strcmp(rss_str, prss_str) != 0)
				{
					start_proc(do_exec, NULL);
				}
			}

			myfree(prss_str, "previous rss string");
			myfree(pnag1_str, "previous nagios string");

			free_stats_array(pstats, n_stats);

			last_file_change = cur_last_file_change;
		}

		if (rss_fd != -1 && FD_ISSET(rss_fd, &rfds) && rss_str != NULL)
		{
			do_client(rss_fd, 1, rss_str, last_file_change, rss_logfile);
		}

		if (nag1_fd != -1 && FD_ISSET(nag1_fd, &rfds) && nag1_str != NULL)
		{
			do_client(nag1_fd, 0, nag1_str, last_file_change, NULL);
		}
	}

	return 0;
}
